#ifndef LABEL_H
#define LABEL_H

#include "typedef.h"

/*{{{  constant definitions*/

/*{{{  flags for 'lp->type' ('LB_...')*/
/*--------------------------------------------------------------*/
/* lp->type  (type information for labelled statements):	*/
/* -----------------------------------------------------	*/
/*	The format of this word is:				*/
/*								*/
/*	lp->type  =  (type_flag | 2nd field -- see below)	*/
/*								*/
/* The 'LB_NTYPES' most significant bits are reserved for the	*/
/* 'type_flag'.  The format of the '2nd field' depends on the	*/
/* 'type flag':							*/
/*  ----------------------------------------------------------	*/
/*	type flag		|	2nd field		*/
/*  ----------------------------|-----------------------------	*/
/*  LB_TARGET_OR_DO_TERM,	|				*/
/*	LB_TARGET or LB_END_IF	| block no. of the stmt.	*/
/*  LB_ILLEGAL_TARGET		| LB_CASE, LB_ELSE, LB_ELSE_IF	*/
/*				|  or empty.			*/
/*  LB_FORMAT			| empty.			*/
/*  ----------------------------------------------------------	*/
/*								*/
/*	'LB_ILLEGAL_TARGET' means that the label cannot be a	*/
/* GOTO target.  Then, either the stmt is executable but not a	*/
/* legal target (viz. ELSE, ELSEIF or CASE), in which case the	*/
/* type of stmt is indicated, or (stmt_part != EXEC_PART).	*/
/*								*/
/*	The 'type flags' are...					*/
/*--------------------------------------------------------------*/

#define  LB_TARGET		0x8000000
#define  LB_TARGET_OR_DO_TERM	0x10000000
#define  LB_END_IF		0x20000000
#define  LB_FORMAT		0x40000000
#define  LB_ILLEGAL_TARGET	0x80000000

#define  LB_DO_TERM		LB_TARGET_OR_DO_TERM
				/* Used only by 'get_newlabel()' */

/*--------------------------------------------------------------*/
/* If the 'type flag' is 'LB_ILLEGAL_TARGET', one of the	*/
/* following flags may be set to specify a particular type of	*/
/* executable stmt that is an illegal target for a GOTO.  These	*/
/* flags occupy bit positions after those of the 'type flag'.	*/
/*--------------------------------------------------------------*/
#define  LB_CASE		0x1000000
#define  LB_ELSE		0x2000000
#define  LB_ELSE_IF		0x4000000

#define  LB_ILLEGAL_CASE	0x81000000  /* LB_ILLEGAL_TARGET | LB_CASE */
#define  LB_ILLEGAL_ELSE	0x82000000  /* LB_ILLEGAL_TARGET | LB_ELSE */
#define  LB_ILLEGAL_ELSE_IF	0x84000000  /* LB_ILLEGAL_TARGET | LB_ELSE_IF */

/*}}}*/

/*{{{  flags for 'refp->type' ('LB_...')*/
/*--------------------------------------------------------------*/
/* refp->type:							*/
/* -----------							*/
/* The format of this word is:					*/
/*								*/
/*	refp->type  =  (type_flag | block no. of the reference)	*/
/*								*/
/* N.B.: It is only useful to record the block number for	*/
/* 'LB_DO_REF's and 'LB_GO_TO_REF's, but for simplicity we	*/
/* might as well always do it (after all, there is no other	*/
/* information to record in the other cases!!)			*/
/*--------------------------------------------------------------*/

/*------------------------------------------------------*/
/* Type flags for label references (refp->type):	*/
/*------------------------------------------------------*/
#define  LB_DO_REF	0x10000000
#define  LB_GO_TO_REF	0x20000000
#define  LB_FORMAT_REF	0x40000000
#define  LB_ASSIGN_REF	0x80000000

/*}}}*/

#define  LB_TYPE_MASK	0xF8000000	/* for both 'lp->type' & 'refp->type' */
#define  LB_BLOCK_MASK	0x07FFFFFF	/* ~LB_TYPE_MASK */


/*{{{  special value of 'lp->label' ('LB_UNLABELLED_DO')*/

/*-------------------------------*/
/* Special values of 'lp->label' */
/*-------------------------------*/
#define  LB_UNLABELLED_DO  -1
/*}}}*/

/*{{{  flags for 'blockp->type' ('BL_...')*/
/*--------------------------------------------------------------*/
/* The 'blockp->type' word holds a keyword token (i.e. TK_xxx)	*/
/* denoting the type of block construct.  The following flags	*/
/* occupy the high-order bits of 'blockp->type', and are only	*/
/* set for DO-loops that contain CYCLE or EXIT statements, or a	*/
/* GO_TO that branches to the terminal statement of the loop.	*/
/*--------------------------------------------------------------*/

#define  BL_CYCLE_REF	0x40000000
#define  BL_EXIT_REF	0x80000000

#define  BL_FLAGS	0xC0000000
#define  BL_TYPE_MASK	~BL_FLAGS




/*}}}*/


#define  BLOCKS_PER_REC		8
#define  BLOCKS_PER_STACK_REC	8	/* adequate for most programs --
					allows blocks nested 8 deep! */

/*{{{  Values of the 'set_label' param of 'lookup_or_install_label()'*/
/*----------------------------------------------------------------*/
/* Values of the 'set_label' param of 'lookup_or_install_label()' */
/*----------------------------------------------------------------*/
#define  SET_LABEL	TRUE
#define  REF_LABEL	FALSE
/*}}}*/


#define  NULL_Block		((Block *) NULL)
#define  NULL_Label		((Label *) NULL)
/*}}}*/
/*{{{  typedefs (Labelref, Label, Block, Blockrec, Blockstack)*/
/*------------------------------------------------------*/
/* Labelref -- a record for each reference to a label	*/
/*------------------------------------------------------*/
typedef struct Labelref {
	int  type;		/* label info (see const. defs above) */
	int  lineno;		/* of label ref - for error messages */
	/*------------------------------------------------------*/
	/* the next 2 members are only relevant to DO-loops...	*/
	/*------------------------------------------------------*/
	int  do_newlabel;	/* for DO-loops only */
	int  do_blockno;	/* for DO-blocks only */
	struct Labelref  *prev;
} Labelref;

/*--------------------------------------------------------------*/
/* Label -- a record for each label (incl. all its references).	*/
/*--------------------------------------------------------------*/
typedef struct Label {
	int  label;
	int  newlabel;
	int  type;		/* label info (see const defs above) */
	int  lineno;		/* of stmt label - for error messages */
	struct Labelref  *last_refp;
	struct Label *next;
} Label;

/*--------------------------------------------------------------*/
/* Blockrec -- a list of records for *all* of the blocks used	*/
/* in the current procedure (even those no longer in scope).	*/
/* The 'type' member identifies the type of block by means of a	*/
/* parser token, eg. TK_DO, TK_DO_WHILE, etc.  The highest-order*/
/* bits are used for the flags BL_CYCLE_REF and BL_EXIT_REF,	*/
/* which are only relevant to DO-loops.				*/
/*--------------------------------------------------------------*/
typedef struct Block {
	int    type;
	char  *name;
	int    block_lineno;		/* start of block - for errmsgs */
	int    constr_lineno;		/* start of construct - for errmsgs */
	/*------------------------------------------------------*/
	/* the remaining members are only relevant to particular*/
	/* types of constructs...				*/
	/*------------------------------------------------------*/
	int    do_label;		/* for DO-loops - label;
					   for IF constructs, #ELSEIF blocks */
	struct Labelref  *do_refp;	/* for DO-loops only */
	struct Block  *default_bp;	/*** MAKE THIS AN UNION WITH ABOVE 2!!  ***/
} Block;

typedef struct Blockrec {
	struct Block  block [BLOCKS_PER_REC];
	struct Blockrec  *next;
} Blockrec;

/*--------------------------------------------------------------*/
/* Blockstack -- a stack of the block nos.  CURRENTLY IN SCOPE.	*/
/*--------------------------------------------------------------*/
typedef struct Blockstack {
	int  blockno [BLOCKS_PER_STACK_REC];
	struct Blockstack *next;
} Blockstack;

/*}}}*/
/*{{{  externs*/

extern bool  just_written_adapt_finish;

extern int   stmt_label,  stmt_newlabel;

extern Label  *stmt_lp;   /* pointer to record of the current stmt label */

Block	*block_name_in_scope (),  *innermost_do_loop(),
	*install_cycle_ref (),  *install_exit_ref (),
	*lookup_block (),  *lookup_current_block ();

int      change_newlabel_to_fmt (),  end_if_block (),  get_newlabel (),
         remove_newlabel ();

Label	*lookup_or_install_label (),  *install_label_ref ();
/*}}}*/

#endif
